
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2007 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.0. Last modification 1/03/07
*)
open Types;;
open Globals;;
open Error_handler;;
open Semantic_check;;
open Parser;;
open Lexer;;
open Interface;;
open Instantiation;;
open Intruder;;



(*
   This is the main compiler module. It parses the command line arguments 
   and invokes the corresponding sub-modules
*)


(* Parsing the arguments of the compiler, for detecting options and
   possible mistakes.
*)

let is_input = ref false in
(Arg.parse
   [
    (
     "--nowarnings",
     Arg.Unit
       (fun () ->
    Globals.flag_warnings := false),
     " Do not display warnings"
    );
    (
     "--debug",
     Arg.Int
       (fun i ->
    Globals.debug_level := i),
     "int   Debug level (0-5)"
    );
    (
     "--version",
     Arg.Unit
       (fun () -> 
   print_string " This is hlpsl2scenario version 1.0. \n";
   print_string " Release date : 18 may 2006.  \n";
   print_string " See the AVISPA Project for details. \n";
   exit 0 ),
     " Print version number"
    );
    (
     "--intruder",
     Arg.Unit
       (fun () -> 
   Intruder.intruder_mode := true
   ),
     " Intruder mode : enable"
    );
  ]
   (fun s ->
     if !is_input then
       raise (Arg.Bad "error: multiple input files")
     else
       (is_input := true;
  Globals.input_file := s))
   "Usage: hlpsl2scenario [options] [file.hlpsl]\nOptions:"
);;


(* opening the input file or the stdin if none: *)
let input_channel =
  if (!Globals.input_file = "") then 
    (Globals.flag_stdout := true;
     Globals.flag_split := false;
     stdin)
  else
    try (
      if not !Globals.flag_stdout then
  (if (Filename.check_suffix !Globals.input_file ".hlpsl") then
     Globals.output_file := Filename.basename (Filename.chop_suffix !Globals.input_file ".hlpsl")
         else
     Globals.output_file := Filename.basename !Globals.input_file;
   if (!Globals.output_dir = "") then
     Globals.output_dir := (Utils_simulator.path_span^"/testsuite/results/");
   Globals.tmp_file := (!Globals.output_dir)^(!Globals.output_file)^".if");
      open_in !Globals.input_file)
    with Sys_error(f) ->
      Error_handler.handleSystemError(f)
in
(try
  (*prerr_endline ("%% Translation of "^(!Globals.input_file));*)

  (* calling the lexer *)
  Globals.lexbuf := Lexing.from_channel input_channel;

  (* calling the parser *)
  let specif = Parser.specification Lexer.token !Globals.lexbuf in
  (* semantic check invocation*)
  Semantic_check.check_all specif;

  (* role instanciation *)
  let basic_instances = Instantiation.call_instantiation specif in

  (* IF printing *)
  Interface.print_if basic_instances specif#get_goals;

  (* error handling *)
with
  Parsing.Parse_error ->
    Error_handler.handleParseError();
    exit(-1)
| Lexer.Eof ->
    prerr_endline "%% Fatal error: unexpected end of file!";
    Error_handler.handleParseError();
    exit(-1);
| x ->
    prerr_string "%% ";
    raise x)
;;
