
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2007 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.0. Last modification 1/03/07
*)
 (**
variables.ml
@author Yann GLOUCHE
@version 1.0
@since 12/12/2005
*)
 
 open Utils_simulator;;
 open Automate;;

  (** signatures of the the roles and types of the variables *)
 
  
  (** contains the signature of the roles, and the type of the variables 
  - the key of the hash table is the name of the role or the name of the variable
  - the value is the signature of the role (a list of type), or the type of the variable (a list with one element : a string for describes the type ([type]))
  *)
  let role_signature = (Hashtbl.create 600 : (string, string list) Hashtbl.t);;

  (**
  for define if the unification respects the type of variables
  *)
  let untyped_variable = (ref false : bool Pervasives.ref);;

  (**
  add a type of a variable or a signature of role in the hash table role_signture
  @param role_name is the name of a role, or a name of variable
  @param signature is the signature of a role or a type of a variable
  @return unit
  *)
  let add_type (role_name : string) (signature : string list) : unit = Hashtbl.add role_signature role_name signature;;
 
  add_type "X_X" ["message"];;
  (** return the type of a variable "r"
  @param r is a name of a variable
  @return a type of a variable
  *)
  let get_type (r : string) : string = 
    let t = Hashtbl.find_all role_signature r in
    if(t<>[]) 
    then list2string "" (car t)
    else "";;

 (** return the signature of a role "role_name"
 @param role_name is a name of a role
 @return the signature of a role
 *)
 let get_role_signature (role_name : string) : string list =
   let t = Hashtbl.find_all role_signature role_name in
   if(t<>[])
   then (car t)
   else [];;

(**
add the signature of the roles contains in the !automate variable, in the hash table role_signature
@param automate is the value of the refernce automate
@return unit
*)
  let print_type_in_role_session (automate : (string * int * 'a * 'b) list): unit = 
    List.iter
     (fun (name,id,_,_) -> 
               add_type (get_name_in_session name id) (get_role_signature name))
     automate;;
 

 (** variables : allocation, valeurs *)
  
 (**  contains the value of the variables of a role
 - the key of the hash table is a pair (name_of_role,name_of_variable)
 - the value is the value of the variable in the role "role_name"
 *)
 let var_value = (Hashtbl.create 600 : (string * string, string) Hashtbl.t);;
 (**
 contains the names of the variables for which the value change for each steps of the protocol execution
 *)
 let variable_story = (ref [] : (string * string) list list Pervasives.ref);;
 (**
 contains the name of the variables for which the value change in the current step of the execution of the protocol
 *)
 let variable_current_change = (ref [] : (string * string) list Pervasives.ref);;
(**
test if value is a variable
@param var is maybe a name of a variable
@return true if "var" is a variable, else false
*)
 let is_a_var (var : string) : bool =
	if (var <> "")
	then( 
		let first = (String.sub var 0 1) in
		if ((String.uppercase first) = first && (String.lowercase first) <> first && not(containStr"(" var) && not(containStr"," var)) then true
		else false;
		)
	else false;;
(**
test if value is a value
@param value is maybe a value
@return true if "value" is a value, else false
*)
 let is_a_value (value : string) : bool=
	if (value <> "")
	then( 
		let first = (String.sub value 0 1) in
		if ((String.lowercase first) = first && not(containStr"(" value) && not(containStr"," value)) then true
		else false;
		)
	else false;;

 (**
 test if "var" is a set
 @param var is maybe a set
 @return true if "var" is a set, else false
 *)
  let is_a_set (var : string) : bool = 
	let type_of_var = get_type var in
	("set")=type_of_var  ||
		(let d = decompose type_of_var in
			(d<>[]&&(car d)="set"))
;;
  (**
 test if "var" is a list
 @param var is maybe a list
 @return true if "var" is a list, else false
 *)
  let is_a_list (var : string) : bool = 
	let type_of_var = get_type var in
	("list")=type_of_var  ||
		(let d = decompose type_of_var in
			(d<>[]&&(car d)="list"))
;;
  (**
 test if "var" is an aggregation type
 @param var is maybe an aggregation type
 @return true if "var" is an aggregation type, else false
 *)
 let is_an_aggregation (var : string) : bool = (is_a_set var) || (is_a_list var);;

 (**
 @param var is a variable
 @return true if the variable "var" is present in a message for test her value
 *)
 let is_a_test_var (var : string) : bool = (is_a_var var) && ((String.length var)>5) && (String.sub var 0 5)="Test_";;
   (**
 @param var is a variable
 @return the real name of the test variable "var"
 *)
 let get_name_of_test_var (var : string) : string = if(is_a_test_var var) then String.sub var 5 ((String.length var)-5) else var;;
 
  (**
 @param var is a variable
 @return true if the variable "var" is present in a message with a new value refresh by the HLPSL function new').
 *)
 let is_a_refresh_var (var : string) : bool = (is_a_var var) && (String.length var)>4 &&  ((String.sub var ((String.length var)-4) 4)="_new");;
  (**
 @param var is a variable
 @return the real name of the refresh variable "var"
 *)
 let get_name_of_refresh_var (var : string) : string = if (is_a_refresh_var var) then (String.sub var 0 ((String.length var)-4) ) else var;;
 
 (**
 @param var is a variable
 @return the real name of the variable "var"
 *)
 let get_name_of_var (var : string) : string = (get_name_of_refresh_var (get_name_of_test_var var))

 
 (**
 Record the variables for which the value is changed during the current execution of the protocol. Each pair is (name_of_an_agent,name_of_a_variable)
 @return unit
 *)
 let add_in_variable_change (var : string * string) : unit = 
	variable_current_change:=(var::(!variable_current_change));; 
 (**
 Record the variables for which the value is changed during the execution of the last step of the protocol. Each pair is (name_of_an_agent,name_of_a_variable)
 @return unit
 *)
 let add_in_variable_story (() : unit) : unit = 
	variable_story:=((!variable_current_change)::(!variable_story));
	variable_current_change:=[];;

(**
the variables comme back to their previous values at the previous step of the execution of the protocol
@return unit
*)
 let previous_variable_state (() : unit) : unit =
	match (!variable_story) with [] -> ()
					|(x::ls) -> ((List.iter (fun t -> (Hashtbl.remove var_value t)) (!variable_current_change));
							variable_current_change:=x;
							variable_story:=ls)
;;
 

(**
Give the value of the variable "var" of the agent "name"
@param name is an agent
@param var is a variable of the agent "name"
@return the value of the variable "var" of the agent "name"
*)			
 let get_value_of_var (name : string) (var : string) : string =
 	if(is_a_var var || (name="x" && is_an_aggregation var))
	then(
  		let var_aux = get_name_of_var var in
		let hv = Hashtbl.find_all var_value (name,var_aux) in
 		if(hv=[]) then (var_aux)
		else ((if((get_type (car hv))="") then add_type (car hv) [(get_type var_aux)]);(car hv));
	)
	else var
;; 

 (**
 Record the variables for which the value is changed during a period of the execution of the protocol. Each pair is (name_of_an_agent,name_of_a_variable).
 It is used for compute the incoming events, and come back at the current step of the protocol.
 *)
 let local_variable_change = (ref [] : (string * string) list Pervasives.ref);;

 (**
 give the value or the value of the variable "value" of the agent "name0" to the variable "var" of the agent "name"
 @param name0 an agent
 @param value a value or a variable of the agent "name0"
 @param name an agent
 @param var a variable of the agent "name"
 @return unit
 *)
 let add_var_value (name0 : string) (value : string) (name : string) (var : string) : unit =
	let value_aux = 	if((is_a_var value))
				then (get_value_of_var name0 value)
 				else value in
	if((get_value_of_var name var)<>value_aux)
	then(
		let var_aux = get_name_of_var var in
		let name_aux = (name,var_aux) in
 		add_in_variable_change name_aux;
		local_variable_change:= name_aux::(!local_variable_change);
		(Hashtbl.add var_value name_aux value_aux);
		if((get_type var_aux)="") then Hashtbl.add role_signature value_aux [(get_type var_aux)];
	)
;;
(** test if the type "t1" can be unified with the type "t2"
@param t1 is a type
@param t2 is a type
@return true, if "t1" can be unified with "t2", else false
*)
  let rec compatible (t1 : string) (t2 : string) : bool = 
  let test t1 t2 = (( t2="message" || t1="message" || (t2="function" && t1="hash_func") ||  (t1="function" && t2="hash_func") || t1=t2) && t1<>"" && t2<>"" )
  in
  if (test t1 t2) then true
  else(let decomposet1 = (decompose t1) in
       let decomposet2 = (decompose t2) in
       if(t1="set" && decomposet1=[] && decomposet2<>[] && (car decomposet2)="set") then true else
       if(t1="list" && decomposet1=[] && decomposet2<>[] && (car decomposet2)="list") then true else
       if(decomposet1<>[] && ((List.length decomposet2)=(List.length decomposet1)) &&(car decomposet1)=(car decomposet2))
       then(
       	let rec aux l1 l2 =
	 if(l1=[] && l2=[]) then true
	 else if(compatible (car l1)(car l2)) then (aux (cdr l1)(cdr l2))
	 else false
	in
	( ((car decomposet1)=(car decomposet2))&&(aux (cdr decomposet1) (cdr decomposet2)) );
       )
       else if (decomposet1<>[] && ((List.length decomposet2)<=(List.length decomposet1)) && (car decomposet1)<>"pair" && (car decomposet1)<>"scrypt" && (car decomposet1)<>"crypt") then true 
       else false)
  ;;

 let rec get_type_of_var_in_state (role_name : string) (state : string) (var : string) : string =
  let give var  state =
	let signature = get_role_signature role_name in
	let state_list = cdr(decompose ("x("^state^")")) in
	let rec aux l_state l_signature =
		if(l_state=[] || l_signature=[]) then ""
		else 
		if ((get_name_of_var (car l_state))=var) then (car l_signature)
		else (aux (cdr l_state)(cdr l_signature))
	in
	let res = ref (aux state_list signature) in 
	if((!res)="") 
	then (let tmp=(get_role_signature var) in
	      if(tmp<>[]) then res:=(car(get_role_signature var)));
	(!res)
  in
  let give_of_all_states_of_role var state =
  	let rec aux l_automate =
		match l_automate with [] -> get_value_of_var role_name var
				|((name,id,states,_)::ls) -> 	if((get_name_in_session name id)=role_name)
								then (
									let rec aux2 l_states =
										match l_states with [] -> aux ls
												|((state0,_)::xs) -> 	if(state0<>state)
															then (	let res= give var state0
																in 
																if(res="") 
																then aux2 xs 
																else res
															)
															else aux2 xs;
									in aux2 (!states)
								)
								else aux ls;
  	in 	let res= get_type var in
		if(res="")
		then (
			let res2 = give var state in
			if(res2="")
			then aux (!automate)
			else res2
		)
		else res;
  in
  let var_decompose = (decompose var) in
  if (var_decompose=[])
  then  (	give_of_all_states_of_role (get_name_of_var var) state )
  else 
  let rec print l = match l with [] -> ""
    			| [x] -> (get_type_of_var_in_state role_name state x)
  			| (x::xs) -> (get_type_of_var_in_state role_name state x)^","^(print xs);
  in
  (car var_decompose)^"("^(print (cdr var_decompose))^")"
  ;;
  
  (** aggregation types : lists, sets *) 

 (** 
add the value or tha variable "value" of the role "name0" is contained in the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return unit
*)
 let add_in_set (name0 : string) (value : string) (name : string) (var : string) : unit =
 	if(is_a_set var)
	then( 
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let value_aux = 	let val_to_add = if((is_a_var value))
							then (get_value_of_var name0 value)
 							else value in
					let decomp = decompose old_value_of_set in
					if(decomp<>[] && not(List.mem val_to_add (cdr decomp))) 
					then ("set("^(list2string "," (val_to_add::(cdr decomp)))^")")
					else if(decomp=[]) then ("set("^val_to_add^")")
					else old_value_of_set
		in
		if(old_value_of_set<>value_aux)
		then(
			let name_aux = ("x",set_to_refer) in
 			add_in_variable_change name_aux;
			local_variable_change:= name_aux::(!local_variable_change);
			(Hashtbl.add var_value name_aux value_aux)
		)
	)
;;

(** 
test if the value or tha variable "value" of the role "name0" is contained in the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return a boolean
*)
 let in_set (name0 : string) (value : string) (name : string) (var : string) : bool =
 	if(is_a_set var)
	then(
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let val_to_add = 
			if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value in
		let decomp = decompose old_value_of_set in
		(decomp<>[]) && (List.mem val_to_add (cdr decomp))
	)
	else false
;;

(** 
delete the value or tha variable "value" of the role "name0" is contained in the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return unit
*)
 let delete_in_set (name0 : string) (value : string) (name : string) (var : string) : unit =
 	if(is_a_set var)
	then(
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let value_aux = 	let val_to_add = if((is_a_var value))
							then (get_value_of_var name0 value)
 							else value in
					let decomp = decompose old_value_of_set in
					if(decomp<>[]) then ("set("^(list2string "," (remove val_to_add (cdr decomp)))^")")
					else if(decomp=[]) then ("set()")
					else old_value_of_set;
		in
		if(old_value_of_set<>value_aux)
		then(
			let name_aux = ("x",set_to_refer) in
 			add_in_variable_change name_aux;
			local_variable_change:= name_aux::(!local_variable_change);
			(Hashtbl.add var_value name_aux value_aux)
		)
	)
;;


(** 
add the value or the variable "value" of the role "name0" is contained in the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return unit
*)
 let add_in_list (name0 : string) (value : string) (name : string) (var : string) : unit =
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let value_aux = 	let val_to_add = if((is_a_var value))
							then (get_value_of_var name0 value)
 							else value in
					let decomp = decompose old_value_of_list in
					if(decomp<>[] && not(List.mem val_to_add (cdr decomp))) 
					then ("list("^(list2string "," (val_to_add::(cdr decomp)))^")")
					else if(decomp=[]) then ("list("^val_to_add^")")
					else old_value_of_list
		in
		if(old_value_of_list<>value_aux)
		then(
			let name_aux = ("x",list_to_refer) in
 			add_in_variable_change name_aux;
			local_variable_change:= name_aux::(!local_variable_change);
			(Hashtbl.add var_value name_aux value_aux)
		)
	)
;;

(** 
test if the value or the variable "value" of the role "name0" is contained in the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return a boolean
*)
 let in_list (name0 : string) (value : string) (name : string) (var : string) : bool =
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let val_to_add = 
			if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value in
		let decomp = decompose old_value_of_list in
		(decomp<>[]) && (List.mem val_to_add (cdr decomp))
	)
	else false
;;
(** 
delete the value or the variable "value" of the role "name0" is contained in the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return unit
*)
 let delete_in_list (name0 : string) (value : string) (name : string) (var : string) : unit =
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let value_aux = 	let val_to_add = if((is_a_var value))
							then (get_value_of_var name0 value)
 							else value in
					let decomp = decompose old_value_of_list in
					if(decomp<>[]) then ("list("^(list2string "," (remove val_to_add (cdr decomp)))^")")
					else if(decomp=[]) then ("list()")
					else old_value_of_list;
		in
		if(old_value_of_list<>value_aux)
		then(
			let name_aux = ("x",list_to_refer) in
 			add_in_variable_change name_aux;
			local_variable_change:= name_aux::(!local_variable_change);
			(Hashtbl.add var_value name_aux value_aux)
		)
	)
;;
(** 
add the value or the variable "value" of the role "name0" is contained in "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return unit
*)
 let add_in_aggregation (name0 : string) (value : string) (name : string) (var : string) : unit =
 	add_in_set name0 value name var;
	add_in_list name0 value name var;
 ;;
 (** 
test if the value or the variable "value" of the role "name0" is contained in "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set or list
@return a boolean
*)
 let in_aggregation (name0 : string) (value : string) (name : string) (var : string) : bool =
 	(in_set name0 value name var) || (in_list name0 value name var)
 ;;
 (** 
delete the value or the variable "value" of the role "name0" is contained in "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return unit
*)
 let delete_in_aggregation (name0 : string) (value : string) (name : string) (var : string) : unit =
  	delete_in_set name0 value name var;
	delete_in_list name0 value name var
 ;;

 let remove_local_variable_change (() : unit) : unit =
 	List.iter
	(fun v -> Hashtbl.remove var_value v ; variable_current_change:=(remove_one v (!variable_current_change)))
	(!local_variable_change);
	local_variable_change:=[]
 ;;
 (**
 return the value of the variable "var" of the role "name" at the step "step" of the execution of the protocol
 @param name is the name of the role
 @param var is the name of the variable
 @param step is the step of the execution
 @return tha value of "var"
 *)
 let rec get_value_of_var_at_step (name : string) (var : string) (step : int) : string =
   let var_aux = if (is_a_test_var var) then (get_name_of_test_var var) else var in
   let variable = (name,var_aux) in
   let all_value_of_variable = Hashtbl.find_all var_value variable in
   let rec give_number_of_occur n l res =
     match l with [] -> res
       |(x::ls) ->	if(n=x)
	 then give_number_of_occur n ls (res+1)
	 else give_number_of_occur n ls res
   in
     if(all_value_of_variable=[]) then var_aux
     else (
       let all_variables_story = (!variable_current_change)::(!variable_story) in
       let rec aux l step pos_res =
	 match l with [] -> pos_res
	   |(x::ls) ->
	      if (step<=0)
	      then pos_res
	      else(
		if(List.mem variable x)
		then aux ls (step-1) (pos_res+(give_number_of_occur variable x 0))
		else aux ls (step-1) pos_res;
	      );
       in 
       let pos = ((min (List.length all_value_of_variable) (aux (List.rev all_variables_story) step 0))-1)  in
       let res = List.nth (List.rev all_value_of_variable) (max 0 pos)
       in 
	 if(pos<0) then res
	 else if(is_an_aggregation res)
	 then (res^"->"^(get_value_of_var_at_step "x" res step))
	 else res;
     )
;; 
(**
	id_refresh is used for refresh the value of the variables. It is the consequence of the HLPSL function new().
*)
 let id_refresh : int Pervasives.ref = ref 1 ;;
 (**
	id_refresh_story is the historic of the variable id_refresh after each step of the protocol.
*)
let id_refresh_story : int list Pervasives.ref = ref [] ;;
 (**
 	record the current value of id_refresh in id_refresh_story
 *)
 let next_id_refresh_story () : unit =
 	id_refresh_story := (!id_refresh)::(!id_refresh_story);
 ;;
 (**
 	id_refresh and id_refresh_story back to their values in the previous step of the protocol
 *)
 let previous_id_refresh_story () : unit =
 	if((!id_refresh_story)<>[]) 
	then ( 
		id_refresh := (car !id_refresh_story) ; 
		id_refresh_story:=(cdr (!id_refresh_story)) 
	)
 ;;
(**
 refresh the value of a variable "var" of the agent "name". It is the consequence of the HLPSL function new().
 @param name is an agent
 @param var is a variable of the agent "name"
 return unit
 *)
 let refresh_var (name : string) (var : string) (tag_number : int) : unit =
	let value_of_var = get_value_of_var name var in
	if((is_a_refresh_var var))
	then (  	
		let value_of_var_tmp = 	if(containStr "-" value_of_var) 
					then String.sub value_of_var 0 (indexStr "-" value_of_var) 
					else value_of_var in
		let value = ((value_of_var_tmp)^"-"^(string_of_int !id_refresh)) in
		id_refresh := 1 + !id_refresh;
		if ((get_type value)="") then add_type value [(get_type (get_name_of_var var))];
		add_var_value name value name var;
	)
 ;;
 
  (**
 refresh the values of all the variables of a state "state" of the agent "name". It is the consequence of the HLPSL function new().
 @param name is an agent
 @param state is a state of the agent "name"
 return unit
 *)
 let refresh_var_of_state (name : string) (state : string) (tag_number : int) : unit = 
 	let decomp = cdr(decompose("x("^state^")")) in
	List.iter (fun v -> 	if (is_a_var v) 
				then refresh_var name v tag_number
				else List.iter (fun v2 -> refresh_var name v2 tag_number) (decompose v)
		) decomp
;;
(**
For debugging : display the value of all the variables on the output error
*)
(*  let view_table_var (() : unit) : unit = 
 	Hashtbl.iter
	(fun (a,b) c -> prerr_string (a^","^b^" -> "^(get_value_of_var a b)^"\n"))
	var_value;
	prerr_string "\n------------\n";
        flush stderr;; *)

 let view_table_var (() : unit) : unit = 
 	Hashtbl.iter
	(fun (a,b) c -> print_string (a^","^b^" -> "^(get_value_of_var a b)^"\n"))
	var_value;
	print_string "\n------------\n";
        flush stdout;;
